@@ -13,11 +13,6 @@
              0
              (min (length string) n)))
 
-;;;###autoload
-(defun haskell-is-prefix-of (x y)
-  "Is x string a prefix of y string?"
-  (string= x (substring y 0 (min (length y) (length x)))))
-
 (defun haskell-string ())
 
 (provide 'haskell-string)
